#include "draw.h"
#include "gpu.h"

void record_ui_draw(VkCommandBuffer command_buffer, UIContext* ui_context, double time) {
  UIPushConstant push = {
    .time = (float)time,
    .layer = 0,
  };

  vkCmdBindPipeline(command_buffer, VK_PIPELINE_BIND_POINT_GRAPHICS, ui_context->pipeline.pipeline);
  vkCmdBindDescriptorSets(command_buffer, VK_PIPELINE_BIND_POINT_GRAPHICS, ui_context->pipeline.layout, 0, 1, &ui_context->font_samplers, 0, NULL);
  vkCmdBindDescriptorSets(command_buffer, VK_PIPELINE_BIND_POINT_GRAPHICS, ui_context->pipeline.layout, 1, 1, &ui_context->font_textures, 0, NULL);
  vkCmdBindDescriptorSets(command_buffer, VK_PIPELINE_BIND_POINT_GRAPHICS, ui_context->pipeline.layout, 2, 1, &ui_context->samplers, 0, NULL);
  vkCmdBindDescriptorSets(command_buffer, VK_PIPELINE_BIND_POINT_GRAPHICS, ui_context->pipeline.layout, 3, 1, &ui_context->textures, 0, NULL);
  for(uint32_t i = 0; i < ui_context->max_containers; i++) {
    if(ui_context->containers[i].id != 0x00000000) {
      for(uint32_t j = 0; j < ui_context->containers[i].layer_count; j++) {
        push.layer = ui_context->containers[i].layers[j].address;
        vkCmdPushConstants(command_buffer, ui_context->pipeline.layout, VK_SHADER_STAGE_VERTEX_BIT | VK_SHADER_STAGE_FRAGMENT_BIT, 0, 16, &push);
        vkCmdDrawIndirect(command_buffer, ui_context->containers[i].layers[j].layer, offsetof(GPULayer, draw), 1, 0);
      }
    }
  }
}

VkResult draw_frame(
    RenderContext* context,
    UIContext* ui,
    double time) {
  VkResult result;

  result = vkWaitForFences(context->device, 1, &context->frame[context->current_frame].ready, VK_TRUE, UINT64_MAX);
  if(result != VK_SUCCESS) {
    return result;
  }

  result = vkResetFences(context->device, 1, &context->frame[context->current_frame].ready);
  if(result != VK_SUCCESS) {
    return result;
  }

  uint32_t image_index;
  result = vkAcquireNextImageKHR(context->device, context->swapchain, UINT64_MAX, context->frame[context->current_frame].image, VK_NULL_HANDLE, &image_index);
  if(result != VK_SUCCESS) {
    return result;
  }

  VkCommandBuffer command_buffer = context->swapchain_command_buffers[image_index];
  VK_RESULT(vkResetCommandBuffer(command_buffer, 0));

  VkCommandBufferBeginInfo begin_info = {
    .sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO,
  }; 
  VK_RESULT(vkBeginCommandBuffer(command_buffer, &begin_info));

  VkViewport viewport = {
    .width = context->swapchain_extent.width,
    .height = context->swapchain_extent.height,
    .maxDepth = 1.0f,
    .minDepth = 0.0f,
  };
  vkCmdSetViewport(command_buffer, 0, 1, &viewport);

  VkRect2D scissor = {
    .extent = context->swapchain_extent,
  };
  vkCmdSetScissor(command_buffer, 0, 1, &scissor);

  VkClearValue clear_values[2] = {{.color={{0.0f, 0.0f, 0.0f, 0.0f}}}, {.depthStencil={1.0f, 0.0f}}};

  VkRenderPassBeginInfo render_pass_begin = {
    .sType = VK_STRUCTURE_TYPE_RENDER_PASS_BEGIN_INFO,
    .renderPass = context->render_pass,
    .framebuffer = context->swapchain_framebuffers[image_index],
    .renderArea.offset = {0, 0},
    .renderArea.extent = context->swapchain_extent,
    .clearValueCount = 2,
    .pClearValues = clear_values,
  }; 

  vkCmdBeginRenderPass(command_buffer, &render_pass_begin, VK_SUBPASS_CONTENTS_INLINE);
  // Render World
  vkCmdNextSubpass(command_buffer, VK_SUBPASS_CONTENTS_INLINE);
  // Render UI
  record_ui_draw(command_buffer, ui, time);
  vkCmdEndRenderPass(command_buffer);

  VK_RESULT(vkEndCommandBuffer(command_buffer));

  VkPipelineStageFlags wait_stages[] = {VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT};
  VkSubmitInfo submit_info = {
    .sType = VK_STRUCTURE_TYPE_SUBMIT_INFO,
    .waitSemaphoreCount = 1,
    .pWaitSemaphores = &context->frame[context->current_frame].image,
    .pWaitDstStageMask = wait_stages,
    .commandBufferCount = 1,
    .pCommandBuffers = &context->swapchain_command_buffers[image_index],
    .signalSemaphoreCount = 1,
    .pSignalSemaphores = &context->frame[context->current_frame].render,
  };

  result = vkQueueSubmit(context->graphics_queue.handle, 1, &submit_info, context->frame[context->current_frame].ready);
  if(result != VK_SUCCESS) {
    return result;
  }

  VkPresentInfoKHR present_info = {
    .sType = VK_STRUCTURE_TYPE_PRESENT_INFO_KHR,
    .waitSemaphoreCount = 1,
    .pWaitSemaphores = &context->frame[context->current_frame].render,
    .swapchainCount = 1,
    .pSwapchains = &context->swapchain,
    .pImageIndices = &image_index,
    .pResults = 0,
  };

  result = vkQueuePresentKHR(context->present_queue.handle, &present_info);
  if(result != VK_SUCCESS) {
    return result;
  }

  context->current_frame = (context->current_frame + 1) % MAX_FRAMES_IN_FLIGHT;

  return VK_SUCCESS;
}
